/*
  Note from Brett:

I use Vide 2.6-rc.2 to open the YM files (important because Malban made some changes 
in there to properly read YM2 and YM3 formats), and then select C down by the Source 
Option, and Encoding set to YMSound optimized speed.  Then click Create Source and it 
exports a .h and *.s to your project folder.  You have to rename the constants in there 
to not contain a hyphen or it won't compile.  I just change name-blah_data to name_ 
everywhere in the *.h and *.s

If you'd like this newer version of Vide... if you dare!  It's here:

Download a single "jar" at:
http://vectrex.malban.de/tmp/Vide_Brett.jar

Replace the jar in your Vide application:
... Vide/Vide.App/Contents/Java/Vide.jar

(and rename to Vide.jar)

It is a fun little game so I'm happy to have helped in some way.  I would love to see 
the high score saved to the BIOS high score area... that way VEXTREME can automatically 
save and recall the high score 

 */

extern const void mad_max_enchanted_lands7_data;
struct Song
{
    const void *data;
};
struct Song song;

//=================================================
// MUSIC / EXPLOSION routines
//=================================================
#define SOUND_EXPL_RISE 0
#define SOUND_EXPL_FALL 1
#define SOUND_EXPL_EVEN -1
#define SOUND_VOL_RISE 0
#define SOUND_VOL_FALL 1
#define SOUND_VOL_EVEN -1
struct sound_adsr_table_t
{
    int amplitude[16];  // 32 nibbles amplitude offsets
};
struct sound_twang_table_t
{
    int frequency[8];   // base frequency offsets
};
struct sound_music_t
{
    const struct sound_adsr_table_t* adsr_table;    // pointer to 32 nibble ADSR table
    const struct sound_twang_table_t* twang_table;  // pointer to 8 byte twang table
    unsigned int music_data[];                      // music data bytes
};
struct sound_explosion_t
{
    int enable;             // noise (bits 5-3) and tone channel (bits 2-0) enable (0b00nnnccc),
    int noise;              // noise source sweep, =0 up, >0 down, <0 inhibit
    int volume;             // volume sweep, =0 up, >0 down, <0 inhibit
    unsigned int duration;  // explosion duration, 0x01 longest to 0x80 shortest
};
const struct sound_explosion_t bang =
{
    0b00101010, SOUND_EXPL_RISE, SOUND_VOL_RISE, 4U
};
const struct sound_explosion_t explosion_off =
{
    0b00111111, SOUND_EXPL_RISE, SOUND_VOL_FALL, 128U
};
const unsigned int fade_fatality[16] = {
    0xEF,0xFE,0xEF,0xFE,0xEF,0xFE,0xEF,0xFE,0xEF,0xFE,0xEF,0xFE,0xEF,0xFE,0xEF,0xFE
};
const struct sound_music_t music_off =
{
    .adsr_table = (struct sound_adsr_table_t*) &Vec_ADSR_FADE4,
    .twang_table = (struct sound_twang_table_t*) &Vec_TWANG_VIBE0,
    .music_data = {128,128,1,128}
};
const struct sound_music_t fatality_tune =
{
    .adsr_table = (struct sound_adsr_table_t*) &fade_fatality,
    .twang_table = (struct sound_twang_table_t*) &Vec_TWANG_VIBE0,
    .music_data =
    {
        128 + 2, 128 +   8, 14, 16,
        128 + 0, 128 +   6, 12, 16,
        128 + 8, 128 +  14, 20, 32,
        1, 0x80,
    }
};
const struct sound_music_t* current_music = &music_off;
const struct sound_explosion_t* current_explosion = &explosion_off;
void sound_init()
{
    current_music = &music_off;
    current_explosion = &explosion_off;
}
void stop_music()
{
    Vec_Music_Flag = 0;
    current_music = &music_off;
}
void stop_explosion()
{
    Vec_Expl_Timer = 0;
    Vec_Expl_Flag = 0;
    current_explosion = &explosion_off;
}
void play_music(const struct sound_music_t* music)
{
    // Vec_Music_Flag == 0 indicates that no current sound is playing
    stop_explosion();
    current_music = music;
    Vec_Music_Flag = 1; // start new music
}
void play_explosion(const struct sound_explosion_t* explosion)
{
    // Vec_Expl_Timer == 0 indicates that no current explosion is playing
    Stop_Sound();
    stop_music();
    current_explosion = explosion;
    Vec_Expl_Flag = 128; // start new explosion
}
int pause_ym_sound = 0;
unsigned int explosion_done = 0;
